#ifndef _zetahat_h
#define _zetahat_h

extern int    zetaHatDefaultRows;
extern int    zetaHatDefaultCols;
extern double zetaHatDefaultGmin;
extern double zetaHatDefaultGmax;
extern double zetaHatDefaultXimin;
extern double zetaHatDefaultXimax;

extern double zetaHatDefaultData[];
extern int    zetaHatDefaultValid[];

#endif
